"""
Zip Archive Functions
"""

### INCLUDES ###
import os
import zipfile
import logging


### CONSTANTS ###
## Logger ##
LOGGER = logging.getLogger(__name__)
# LOGGER.setLevel(logging.DEBUG)


### FUNCTIONS ###
def create_zip(zip_path, files_path):
    """ Creates Zip Archive """
    zip_status = True

    try:
        zip_archive = zipfile.ZipFile(zip_path, 'w')

        for root, dirs, files in os.walk(files_path):
            for _file in files:
                zip_archive.write(os.path.join(root, _file))

        zip_archive.close()

    except:
        zip_status = False

    return zip_status


def extract_zip(zip_path, extract_path=None, password=None):
    """ Opens Zip Archive in order to extract files """
    zip_status = True

    try:
        zip_archive = zipfile.ZipFile(zip_path, 'r')
        if password is not None:
            zip_archive.setpassword(password)

        if extract_path is None:
            extract_path = os.path.dirname(zip_path)

        zip_archive.extractall(extract_path)
        zip_archive.close()

    except:
        zip_status = False

    return zip_status



"""
Web Upload Functions
"""

### INCLUDES ###
import os
import codecs
import logging


### CONSTANTS ###
## Read Chunk Size ##
CHUNK_SIZE = 512 * 1024         # 0.5 MB


## Logger ##
LOGGER = logging.getLogger(__name__)
# LOGGER.setLevel(logging.DEBUG)


### FUNCTIONS ###
## Upload related Functions with recursion ##
def save_upload(upload_folder_path, upload_data):
    """ Tries to open Upload File """
    output = False

    upload_path = os.path.join(upload_folder_path, upload_data.filename)
    try:
        upload_file = codecs.open(upload_path, 'w+')
    except:
        pass
    else:
        while True:
            data_chunk = upload_data.file.read(CHUNK_SIZE)

            if not data_chunk:
                break
            else:
                upload_file.write(data_chunk)

        upload_file.close()
        output = upload_path

    return output



"""
Pickle/Unpickle Functions
"""

### INCLUDES ###
import os
import sys
import cPickle
import logging

from .file_system import open_file, write_file


### CONSTANTS ###
## Logger ##
LOGGER = logging.getLogger(__name__)
# LOGGER.setLevel(logging.DEBUG)


### FUNCTIONS ###
## Pickle/Unpickle File ##
def pickle_file(file_path, data_to_pickle):
    """ Pickle file """
    output = False          # Not sure why False, because data can be '' or None?

    file_instance = open_file(file_path, 'w+')
    if file_instance is not None:
        file_name = os.path.basename(file_path)
        try:
            pickled_data = cPickle.dumps(data_to_pickle)
        except KeyboardInterrupt:
            # Retry
            # TODO: Might want to have several retries, not just single retry
            pickle_file(file_path, data_to_pickle)
        except TypeError as error:
            LOGGER.error("Could not pickle " + file_name + "!")
            LOGGER.error("Type Error Details: " + str(error))
        except:
            LOGGER.error("Could not pickle " + file_name + "!")
            LOGGER.error("Pickler error:" + str(sys.exc_info()[0]))
        else:
            output = write_file(file_instance, pickled_data)

        file_instance.close()

    return output


def unpickle_file(file_path):
    """ Unpickle file """
    output = False          # Not sure why False, because data can be '' or None?

    file_instance = open_file(file_path, 'r')
    if file_instance is not None:
        pickled_data = file_instance.read()
        try:
            unpickled_data = cPickle.loads(pickled_data)
        except:
            file_name = os.path.basename(file_path)
            LOGGER.error("Could not unpickle " + file_name + "!")
            LOGGER.error("Unpickler error:" + str(sys.exc_info()[0]))
        else:
            output = unpickled_data

        file_instance.close()

    return output



"""
AES Encrypt/Decrypt Library
"""

### IMPORTS ###
from hashlib import md5

from Crypto.Cipher import AES
from Crypto import Random


### FUNCTIONS ###
def encrypt(in_filename, out_filename, password):
    """ External Encrypt function """
    try:
        in_file = open(in_filename, 'rb')
        out_file = open(out_filename, 'wb')
        in_file.seek(0)
        out_file.seek(0)
        _encrypt(in_file, out_file, password)
    except:
        return False
    else:
        in_file.close()
        out_file.close()
        return True


def _encrypt(in_file, out_file, password, key_length=32):
    """ Internal Encrypt function """
    bs = AES.block_size
    salt = Random.new().read(bs - len('Salted__'))
    key, iv = _derive_key_and_iv(password, salt, key_length, bs)
    cipher = AES.new(key, AES.MODE_CBC, iv)
    out_file.write('Salted__' + salt)
    finished = False
    while not finished:
        chunk = in_file.read(1024 * bs)
        if len(chunk) == 0 or len(chunk) % bs != 0:
            padding_length = bs - (len(chunk) % bs)
            chunk += padding_length * chr(padding_length)
            finished = True
        out_file.write(cipher.encrypt(chunk))


def decrypt(in_filename, out_filename, password):
    """ External Decrypt function """
    try:
        in_file = open(in_filename, 'rb')
        out_file = open(out_filename, 'wb')
        in_file.seek(0)
        out_file.seek(0)
        _decrypt(in_file, out_file, password)
    except:
        return False
    else:
        in_file.close()
        out_file.close()
        return True


def _decrypt(in_file, out_file, password, key_length=32):
    """ Internal Decrypt function """
    bs = AES.block_size
    salt = in_file.read(bs)[len('Salted__'):]
    key, iv = _derive_key_and_iv(password, salt, key_length, bs)
    cipher = AES.new(key, AES.MODE_CBC, iv)
    next_chunk = ''
    finished = False
    while not finished:
        chunk, next_chunk = next_chunk, cipher.decrypt(in_file.read(1024 * bs))
        if len(next_chunk) == 0:
            padding_length = ord(chunk[-1])
            if padding_length < 1 or padding_length > bs:
                raise ValueError("bad decrypt pad (%d)" % padding_length)
            # all the pad-bytes must be the same
            if chunk[-padding_length:] != (padding_length * chr(padding_length)):
                # this is similar to the bad decrypt:evp_enc.c from openssl program
                raise ValueError("bad decrypt")
            chunk = chunk[:-padding_length]
            finished = True
        out_file.write(chunk)


def _derive_key_and_iv(password, salt, key_length, iv_length):
    """ Derive Key and IV """
    d = d_i = ''
    while len(d) < key_length + iv_length:
        d_i = md5(d_i + password + salt).digest()
        d += d_i
    return d[:key_length], d[key_length:key_length + iv_length]



"""
This module functionality consist of:

    * File Manipulation Functions
    * Directory Manipulation Functions
    * File/Folder Memory Space/Size Functions
    * Open/Write/Read File Functions
    * Time Stamp Functions

"""

### INCLUDES ###
import os
import sys
import shutil
import glob
import ctypes
import codecs
import time
import logging

from distutils.spawn import find_executable


### CONSTANTS ###
## Logger ##
LOGGER = logging.getLogger(__name__)
# LOGGER.setLevel(logging.DEBUG)


### FUNCTIONS ###
## File Functions ##
def make_file(file_path):
    """ Creating empty file """
    if not os.path.isfile(file_path):
        LOGGER.debug('Creating file: ' + str(file_path))
        new_file = open(file_path, 'a')
        # Clear content of the file (just in case)
        # new_file.seek(0)
        # new_file.truncate()
        new_file.close()


def remove_file(file_path):
    if os.path.isfile(file_path):
        LOGGER.debug('Removing file: ' + str(file_path))
        os.remove(file_path)


def remove_files(files_path, exclude_paths=None):
    """ Removes files, specified by path/filter """
    if exclude_paths is None:
        exclude_paths = list()

    files = glob.glob(files_path)
    # Remove files first
    for item in files:
        if item not in exclude_paths:
            remove_file(item)

    # Remove directories second
    for item in files:
        if item not in exclude_paths:
            remove_dir(item)


def copy_file(source_file, destination_file, permissions=None, dos2unix=True):
    """
    Copying File
    :param source_file: Source File Path
    :param destination_file: Destination File Path
    :param permissions: Permissions string (for unix only)
    :param dos2unix: Conversion to unix format (for unix only)
    :return: True if copying is successful, False otherwise
    """
    output = os.path.isfile(source_file)
    if output:
        # Remove destination file (if exists)
        if os.path.isfile(destination_file):
            os.remove(destination_file)

        # Copy file
        shutil.copy(source_file, destination_file)

        if os.name == 'posix':
            # Convert to unix format
            if dos2unix:
                if find_executable('dos2unix'):
                    os.system('dos2unix ' + destination_file)

            # Add permissions
            if permissions is not None:
                os.system('chmod ' + str(permissions) + ' ' + destination_file)

    return output


def fetch_file(file_path):
    """ Fetches files, specified by path/filter """
    output = None

    files = glob.glob(file_path)
    if len(files):
        csv_path = files[0]
        # Provide with csv name as an output
        output = os.path.basename(csv_path)

    return output


## Directory Functions ##
def make_dir(dir_path):
    """
    Creating directory
    :param dir_path: Path of the new directory
    :return: True if folder has been created, False if folder exists already
    """
    output = not os.path.isdir(dir_path)
    if output:
        LOGGER.debug('Creating folder: ' + str(dir_path))
        os.makedirs(dir_path)

    return output


def remove_dir(dir_path):
    """
    Removing directory
    :param dir_path: Path of the directory
    :return: True if remove successful (such directory existed in the first place), False otherwise
    """
    output = os.path.isdir(dir_path)
    if output:
        LOGGER.debug('Removing folder: ' + str(dir_path))
        shutil.rmtree(dir_path)

    return output


def empty_dir(dir_path):
    """
    Empty directory
    :param dir_path: Path of the directory
    :return: True if empty successful (such directory existed in the first place), False otherwise
    """
    output = remove_dir(dir_path)
    if output:
        os.makedirs(dir_path)

    return output


def copy_dir(source_path, destination_path):
    # LOGGER.debug('source: ' + str(source_path))
    # LOGGER.debug('destination: ' + str(destination_path))
    if os.path.isdir(source_path):
        make_dir(destination_path)
        sub_items = glob.glob(os.path.join(source_path, '*'))
        for sub_item_path in sub_items:
            sub_item_name = os.path.basename(sub_item_path)
            copy_dir(sub_item_path, os.path.join(destination_path, sub_item_name))

    elif os.path.isfile(source_path):
        copy_file(source_path, destination_path)


## File/Folder Memory Space/Size Functions ##
def get_size(path):
    """
    Taken from here
    http://stackoverflow.com/questions/1392413/calculating-a-directory-size-using-python
    """
    total_size = 0

    if os.path.isfile(path):
        try:
            stat = os.stat(path)
        except OSError:
            pass
        else:
            total_size += stat.st_size

    elif os.path.isdir(path):
        seen = set()
        for directory_path, directory_names, file_names in os.walk(path):
            for f in file_names:
                fp = os.path.join(directory_path, f)

                try:
                    stat = os.stat(fp)
                except OSError:
                    continue

                if stat.st_ino in seen:
                    continue

                seen.add(stat.st_ino)

                total_size += stat.st_size

    return total_size  # size in bytes


def get_free_space(path):
    """
    Return free space in bytes
    Taken from here, slightly modified:
    http://stackoverflow.com/questions/51658/cross-platform-space-remaining-on-volume-using-python
    """
    if os.name == 'nt':
        free_bytes = ctypes.c_ulonglong(0)
        ctypes.windll.kernel32.GetDiskFreeSpaceExW(ctypes.c_wchar_p(path), None, None, ctypes.pointer(free_bytes))
        return free_bytes.value
    else:
        # TODO: Works on Linux, double check this works on MAC, consider following link:
        # http://ilostmynotes.blogspot.com/2014/05/cross-platform-python-method-for.html
        st = os.statvfs(path)
        return st.f_bavail * st.f_frsize


def print_memory_size(size):
    """
    Taken from here
    http://stackoverflow.com/questions/1392413/calculating-a-directory-size-using-python
    """
    _bytes = 'B'
    _kilo_bytes = 'KB'
    _mega_bytes = 'MB'
    _giga_bytes = 'GB'
    _tera_bytes = 'TB'
    memory_units = [_bytes, _kilo_bytes, _mega_bytes, _giga_bytes, _tera_bytes]
    memory_format = '%f %s'
    memory_radix = 1024.

    for u in memory_units[:-1]:
        if size < memory_radix:
            return memory_format % (size, u)
        size /= memory_radix

    return memory_format % (size, memory_units[-1])


## Open/Write/Read File Functions ##
def open_file(file_path, mode, encoding=None):
    """ Tries reading file. Returns false if failed """
    output = None
    try:
        file_instance = codecs.open(file_path, mode, encoding)
    except:
        pass
    else:
        output = file_instance

    return output


def save_file(file_path, file_content, encoding=None, permissions=None):
    """ Tries to open file for writing """
    output = False

    file_instance = open_file(file_path, 'w+', encoding)
    if file_instance is not None:
        if write_file(file_instance, file_content):
            output = os.path.basename(file_path)
        file_instance.close()

        if os.name == 'posix':
            # Add permissions
            if permissions is not None:
                os.system('chmod ' + str(permissions) + ' ' + file_path)

    return output


def write_file(file_instance, file_content):
    """ Tries writing file """
    output = False
    try:
        file_instance.write(file_content)
    except:
        LOGGER.error("write_file unexpected error:" + str(sys.exc_info()[0]))
    else:
        output = True

    return output


## Time Stamp Functions ##
def create_time_stamp(time_stamp_format):
    return time.strftime(time_stamp_format, time.localtime(time.time()))



# !/usr/bin/env python
"""
Python Distribution Package Utilities

Used quite a few recipes from here:
https://wiki.python.org/moin/Distutils/Cookbook
Some hints on how to extend commands
http://stackoverflow.com/questions/1321270/how-to-extend-distutils-with-a-simple-post-install-script/1321345#1321345
"""


### INCLUDES ###
import os
import imp
import glob
import shutil
import logging

from .ordered_dict import OrderedDict
from . import file_system


### CONSTANTS ###
## Logger ##
LOGGER = logging.getLogger(__name__)
# LOGGER.setLevel(logging.DEBUG)


### FUNCTIONS ###
def is_package(path):
    return os.path.isfile(os.path.join(path, '__init__.py'))


def find_packages(path, base=""):
    """ Find all packages in path """
    py_packages = OrderedDict()

    for item in os.listdir(path):
        py_dir = os.path.join(path, item)
        if os.path.isdir(py_dir) and is_package(py_dir):
            # print "item = ", item
            if base:
                module_name = "%(base)s.%(item)s" % vars()
            else:
                module_name = item
            py_packages[module_name] = py_dir
            py_packages.update(find_packages(py_dir, module_name))

    return py_packages


def non_python_files(path, ignore_dirs=None):
    """ Return all non-python-file file names in path """
    result = []
    all_results = []
    module_suffixes = [info[0] for info in imp.get_suffixes()]

    if ignore_dirs is None:
        ignore_dirs = []

    if os.path.isdir(path):
        for item in os.listdir(path):
            name = os.path.join(path, item)
            if os.path.isfile(name) and os.path.splitext(item)[1] not in module_suffixes:
                result.append(name)
            elif os.path.isdir(name) and item.lower() not in ignore_dirs:
                all_results.extend(non_python_files(name, ignore_dirs))
        if result:
            all_results.append((path, result))

    return all_results


def package_data_files(path, ignore_dirs=None):
    """ Returns all file names in path in package data format """
    result = []
    for item in non_python_files(path, ignore_dirs):
        result += item[1]

    return result


def generate_docs(doc_packages):
    """ Generates documentation. Performed before generating distribution on host (Windows) system """
    if len(doc_packages):
        _base_package_names = list()

        # Rebuilding package module interconnections and package classes images
        for _package_name, _package_path in doc_packages.items():
            # Figuring out base package name
            _base_package_name = _package_name.split('.')[0]
            if _base_package_name not in _base_package_names:
                _base_package_names.append(_base_package_name)

            # Rebuild images
            _package_path = _package_path.split(_base_package_name)
            _package_path[0] = ''
            _package_path = _base_package_name.join(_package_path)
            pyreverse_command = 'pyreverse -o png -p ' + _package_name + ' ' + _package_path
            print pyreverse_command
            os.system(pyreverse_command)

        # Moving images to appropriate folder
        images = glob.glob('*.png')
        destination_path = os.path.join('_docs', 'images')
        file_system.make_dir(destination_path)
        for image_path in images:
            image_name = os.path.basename(image_path)
            shutil.move(image_path, os.path.join(destination_path, image_name))

        # Updating automatically generated rst files
        for _base_package_name in _base_package_names:
            apidoc_str = 'sphinx-apidoc -f -o _docs ' + _base_package_name
            print apidoc_str
            os.system(apidoc_str)

        # Rebuilding documentation in html format
        file_system.remove_dir('docs')
        build_str = 'sphinx-build -b html _docs docs'
        print build_str
        os.system(build_str)



# !/usr/bin/env python
"""
Swiss Army Knife Setup Scripts
"""


### INCLUDES ###
import sys

from distutils.core import setup

from py_knife import __version__
from py_knife.py_setup import find_packages, non_python_files, package_data_files, generate_docs


### SETUP PROCEDURES ###
packages = find_packages(".", "")
# print "packages = ", str(packages), "\n"

if len(sys.argv):
    if 'sdist' in sys.argv:
        print "*** Generating Documentation ***"
        generate_docs(packages)

        print "*** Generation Distribution ***"
    elif 'install' in sys.argv:
        # Enable force to overwrite existing files and create folders
        if '--force' not in sys.argv:
            sys.argv.append('--force')

# Probably none, kept for future reference
data_files = (non_python_files('py_knife'))
# print "data_files = ", str(data_files), "\n"

package_data_content = package_data_files('docs')
package_data = {'': package_data_content}
# print "package_data = ", str(package_data), "\n"

setup(
    name='py_knife',
    version=__version__,
    description='Swiss Army Knife',
    long_description='Swiss Army Knife of Python',
    author='Kirill V. Belyayev',
    author_email='kbelyayev@gmail.com',
    url='https://github.com/Barmaley13/SwissArmyKnife',
    # download_url='https://github.com/Barmaley13/SwissArmyKnife/tarball/' + __version__,
    packages=packages.keys(),
    package_dir=packages,
    package_data=package_data,
    data_files=data_files,
    requires=[
        'pyserial',
        'pycrypto'
    ]
)



"""
Python Distribution Install Functions
"""

### INCLUDES ###
import os
import subprocess
import logging

from .file_system import remove_dir


### CONSTANTS ###
## Logger ##
LOGGER = logging.getLogger(__name__)
# LOGGER.setLevel(logging.DEBUG)


### FUNCTIONS ###
def install_package(package_name, package_path=None):
    """ Runs install inside of install """
    cwd = os.getcwd()

    clean_up = False
    if package_path is None:
        package_path = os.path.join(cwd, package_name)
        clean_up = True
    os.chdir(package_path)

    print 'Installing ' + package_name + '...'
    if os.name == 'nt':
        # This failed on E10 as part of web interface install
        install_process = subprocess.Popen(['python', 'setup.py', 'install'], shell=True)
    else:
        # Proven to work robust on E10 (but fails on Windows)
        install_process = subprocess.Popen(['python setup.py install'], shell=True)

    install_process.wait()
    os.chdir(cwd)

    if clean_up:
        print 'Cleaning ' + package_name + ' up...'
        remove_dir(package_path)
